"""Adult Data Set - UCI Machine Learning Repository."""

from typing import Dict, List, Optional

import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.preprocessing import MinMaxScaler

from canonical_sets.data.base import BaseData

# names of the columns in the downloaded csv file
NAMES = [
    "Age",
    "Workclass",
    "fnlwgt",
    "Education",
    "Education-Num",
    "Martial Status",
    "Occupation",
    "Relationship",
    "Race",
    "Sex",
    "Capital Gain",
    "Capital Loss",
    "Hours per week",
    "Country",
    "Target",
]


class Adult(BaseData):
    """Adult Data Set - UCI Machine Learning Repository.

    This class downloads and preprocesses the Adult dataset as
    a `pd.DataFrame`.

    Attributes
    ----------
    train_data : pd.DataFrame
        The training data.
    test_data : pd.DataFrame
        The testing data.
    train_labels : pd.DataFrame
        The training labels.
    test_labels : pd.DataFrame
        The testing labels.
    val_data : pd.DataFrame
        The validation data.
    val_labels : pd.DataFrame
        The validation labels.
    numerical_cols : List[str]
        The numerical columns.
    categorical_cols : List[str]
        The categorical columns.

    Example
    -------
    >>> adult = Adult()
    """

    train_data: pd.DataFrame
    val_data: Optional[pd.DataFrame]
    test_data: pd.DataFrame
    train_labels: Optional[pd.DataFrame]
    val_labels: Optional[pd.DataFrame]
    test_labels: Optional[pd.DataFrame]
    numerical_cols: Optional[List[str]]
    categorical_cols: Optional[List[str]]

    def __init__(
        self,
        train_path: Optional[str] = None,
        test_path: Optional[str] = None,
        download_train_path: Optional[str] = None,
        download_test_path: Optional[str] = None,
        features: Optional[List[str]] = None,
        groups: Optional[Dict[str, Dict[str, str]]] = None,
        scaler: TransformerMixin = MinMaxScaler(feature_range=(-1, 1)),
        prefix_sep: str = "+",
        val_prop: float = 0.2,
        preprocess: bool = True,
        seed: int = 1234,
    ):
        """Initialize the data.

        Parameters
        ----------
        train_path : str, optional
            The path to the training data if it is already downloaded.
        test_path : str, optional
            The path to the testing data if it is already downloaded.
        download_train_path : str, optional
            The path to save the training data to (needs to end in .csv).
            The default is ``None``.
        download_test_path : str, optional
            The path to save the testing data to (needs to end in .csv).
            The default is ``None``.
        features: List[str], optional
            The features to use. The default is ``None``.
        groups: Dict[str, Dict[str, str]], optional
            The groups to use. The default is ``None``.
        scaler : sklearn.base.TransformerMixin
            Any of the ``sklearn`` preprocessing modules.
            The default is ``sklearn.preprocessing.MinMaxScaler``.
        prefix_sep : str
            The prefix separator to split the categorical feature and category
            when one-hot encoding. For example, Color = [Red, Green] ->
            Color+Red and Color+Green. The default is ``+``.
        val_prop: float
            The proportion of the training data to use for validation.
            The default is 0.2.
        preprocess: bool
            Whether to preprocess the data. The default is ``True``.
        seed: int
            The seed for the random state. The default is 1234.
        """
        super().__init__(
            features, groups, scaler, prefix_sep, val_prop, 0, preprocess, seed
        )

        if self.features:
            self.features = features
        else:
            self.features = [
                "Age",
                "Workclass",
                "fnlwgt",
                "Education",
                "Education-Num",
                "Martial Status",
                "Occupation",
                "Relationship",
                "Race",
                "Sex",
                "Capital Gain",
                "Capital Loss",
                "Hours per week",
                "Country",
            ]

        if train_path and test_path:
            train_data = pd.read_csv(train_path)
            test_data = pd.read_csv(test_path)

        else:
            train_url = (
                "https://archive.ics.uci.edu/ml/"
                "machine-learning-databases/adult/adult.data"
            )
            test_url = (
                "https://archive.ics.uci.edu/ml/"
                "machine-learning-databases/adult/adult.test"
            )

            train_data = pd.read_csv(
                train_url,
                names=NAMES,
                sep=r"\s*,\s*",
                engine="python",
                na_values="?",
                header=None,
            )

            test_data = pd.read_csv(
                test_url,
                names=NAMES,
                sep=r"\s*,\s*",
                engine="python",
                na_values="?",
                header=None,
                skiprows=1,
            )

            if download_train_path:
                train_data.to_csv(download_train_path, index=False)

            if download_test_path:
                test_data.to_csv(download_test_path, index=False)

        if preprocess:
            self._preprocess(train_data, test_data)

        else:
            self.train_data = train_data
            self.test_data = test_data

            self.val_data = None
            self.train_labels = None
            self.val_labels = None
            self.test_labels = None
            self.numerical_cols = None
            self.categorical_cols = None

    def _preprocess(
        self, train_data: pd.DataFrame, test_data: pd.DataFrame
    ) -> None:
        """Preprocess the data.

        Parameters
        ----------
        train_data: pd.DataFrame
            The training data.
        test_data: pd.DataFrame
            The testing data.

        Returns
        -------
        None.
        """

        # drop NA
        train_data.dropna(inplace=True)
        test_data.dropna(inplace=True)

        train_data.reset_index(drop=True, inplace=True)
        test_data.reset_index(drop=True, inplace=True)

        df = pd.concat([train_data, test_data])

        # get labels, drop target (erase unnecessary "." in test data)
        df["Target"] = df["Target"].str.replace(r".", "", regex=True)

        labels = pd.get_dummies(df["Target"])
        df.drop("Target", inplace=True, axis=1)

        # drop columns
        df = df[self.features]

        # create groups
        # if self.groups is None:
        #     others = list(df.Country.unique())
        #     others.remove("United-States")
        #     self.groups = {"Country": dict.fromkeys(others, "Others")}

        if self.groups is not None:
            df = self._create_groups(df, self.groups)

        # split data
        x_train = df.iloc[: len(train_data)]
        y_train = labels.iloc[: len(train_data)]

        x_test = df.iloc[len(train_data) :]
        y_test = labels.iloc[len(train_data) :]

        if self.val_prop > 0:
            (
                x_train,
                x_val,
                y_train,
                y_val,
            ) = self._split_data(x_train, y_train, self.val_prop)

            data = pd.concat([x_train, x_val, x_test])

        else:
            data = pd.concat([x_train, x_test])

        # one-hot encode categorical columns
        dummies = self._one_hot_encode(data)

        # scale numerical columns
        scaled_train_data = self._scale_numeric(x_train, fit_scaler=True)
        scaled_test_data = self._scale_numeric(x_test)

        if self.val_prop > 0:
            scaled_val_data = self._scale_numeric(x_val)

        # merge the pre-processed data and attribute to self
        self.train_data = scaled_train_data.join(dummies[: len(x_train)])
        self.train_labels = y_train

        self.test_labels = y_test

        if self.val_prop > 0:
            self.val_data = scaled_val_data.join(
                dummies[len(x_train) : (len(x_train) + len(x_val))]
            )
            self.test_data = scaled_test_data.join(
                dummies[(len(x_train) + len(x_val)) :]
            )

            self.val_labels = y_val

        else:
            self.test_data = scaled_test_data.join(dummies[len(x_train) :])



"""Base class for data sets."""

from typing import Dict, List, Optional, Tuple

import joblib
import pandas as pd
import torch
from sklearn.base import TransformerMixin
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from torch.utils.data import Dataset


class BaseData:
    """Base class for data sets.

    This is a base class from which all data sets inherit.

    Attributes
    ----------
    train_data : pd.DataFrame
        The training data.
    test_data : pd.DataFrame
        The testing data.
    train_labels : pd.DataFrame
        The training labels.
    test_labels : pd.DataFrame
        The testing labels.
    val_data : pd.DataFrame
        The validation data.
    val_labels : pd.DataFrame
        The validation labels.
    numerical_cols : List[str]
        The numerical columns.
    categorical_cols : List[str]
        The categorical columns.
    """

    train_data: pd.DataFrame
    val_data: Optional[pd.DataFrame]
    test_data: Optional[pd.DataFrame]
    train_labels: pd.DataFrame
    val_labels: Optional[pd.DataFrame]
    test_labels: Optional[pd.DataFrame]
    numerical_cols: Optional[List[str]]
    categorical_cols: Optional[List[str]]

    def __init__(
        self,
        features: Optional[List[str]] = None,
        groups: Optional[Dict[str, Dict[str, str]]] = None,
        scaler: TransformerMixin = MinMaxScaler(feature_range=(-1, 1)),
        prefix_sep: str = "+",
        val_prop: float = 0.2,
        test_prop: float = 0.2,
        preprocess: bool = True,
        seed: int = 1234,
    ):
        """Initialize the data.

        Parameters
        ----------
        features: List[str], optional
            The features to use. The default is ``None``.
        groups: Dict[str, Dict[str, str]], optional
            The groups to use. The default is ``None``.
        scaler : sklearn.base.TransformerMixin
            Any of the ``sklearn`` preprocessing modules for the numerical
            features. The default is ``sklearn.preprocessing.MinMaxScaler``.
        prefix_sep : str
            The prefix separator to split the categorical feature and category
            when one-hot encoding. For example, Color = [Red, Green] ->
            Color+Red and Color+Green. The default is ``+``.
        val_prop: float
            The proportion of the training data to use for validation.
            The default is 0.2.
        test_prop: float
            The proportion of the training data to use for testing.
             The default is 0.2.
        preprocess: bool
            Whether to preprocess the data. The default is ``True``.
        seed: int
            The seed for the random state. The default is 1234.

        Raises
        ------
        ValueError
            Proportions must be between [0, 1).
        """
        if val_prop >= 1 or test_prop >= 1 or val_prop < 0 or test_prop < 0:
            raise ValueError("Proportions must be between [0, 1).")

        self.features = features
        self.groups = groups
        self.scaler = scaler
        self.prefix_sep = prefix_sep
        self.val_prop = val_prop
        self.test_prop = test_prop
        self.preprocess = preprocess
        self.seed = seed

        self.train_data = pd.DataFrame()
        self.val_data = None
        self.test_data = None

        self.train_labels = pd.DataFrame()
        self.val_labels = None
        self.test_labels = None

        self.numerical_cols = None
        self.categorical_cols = None

    @classmethod
    def load(cls, path):
        """Load the data.

        Parameters
        ----------
        path: str
            The path to load the data from (needs to end in .pkl).
        """
        return joblib.load(path)

    def save(self, path: str) -> None:
        """Save the object.

        Parameters
        ----------
        path: str
            The path to save the object (needs to end in .pkl).
        """
        joblib.dump(self, path)

    def inverse_preprocess(self, data: pd.DataFrame) -> pd.DataFrame:
        """Inverse preprocess the data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to inverse preprocess.

        Returns
        -------
        pd.DataFrame
            The inverse preprocessed data.
        """
        numbers = self._inverse_scale_numeric(data)
        dummies = self._inverse_ohe(data)

        return numbers.join(dummies)

    def _create_groups(
        self, data: pd.DataFrame, groups: Dict[str, Dict[str, str]]
    ) -> pd.DataFrame:
        """Create groups for the data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to group.
        groups: Dict[str, Dict[str, str]]
            The groups to use. The default is ``None``.

        Returns
        -------
        df: pd.DataFrame
            The grouped data.
        """
        df = data.replace(groups)

        return df

    def _split_data(
        self,
        data: pd.DataFrame,
        labels: pd.DataFrame,
        prop: float = 0.2,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Split the data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to split.
        labels: pd.DataFrame
            The labels to split.
        prop: int
            The proportion of the data to use. The default is 0.2.

        Returns
        -------
        x_data: pd.DataFrame
            The x data.
        y_data: pd.DataFrame
            The y data.
        x_labels: pd.DataFrame
            The x labels.
        y_labels: pd.DataFrame
            The y labels.
        """
        x_data, y_data, x_labels, y_labels = train_test_split(
            data, labels, test_size=prop, random_state=self.seed
        )

        x_data.reset_index(drop=True, inplace=True)
        y_data.reset_index(drop=True, inplace=True)

        x_labels.reset_index(drop=True, inplace=True)
        y_labels.reset_index(drop=True, inplace=True)

        return x_data, y_data, x_labels, y_labels

    def _one_hot_encode(self, data: pd.DataFrame) -> pd.DataFrame:
        """One-hot encode the data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to one-hot encode.

        Returns
        -------
        data: pd.DataFrame
            The one-hot encoded data.
        """
        categories = data.select_dtypes(exclude="number")

        if categories.empty:
            return pd.DataFrame(index=data.index)

        self.categorical_cols = categories.columns.to_list()
        dummies = pd.get_dummies(
            categories,
            prefix=self.categorical_cols,
            prefix_sep=self.prefix_sep,
        )

        return dummies

    def _scale_numeric(
        self, data: pd.DataFrame, fit_scaler: bool = False
    ) -> pd.DataFrame:
        """Scale the numeric data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to scale.
        fit_scaler: bool
            Whether to fit the scaler. The default is ``False``.

        Returns
        -------
        data: pd.DataFrame
            The scaled data.
        """
        numerical_data = data.select_dtypes(include="number")

        if numerical_data.empty:
            return pd.DataFrame(index=data.index)

        if fit_scaler:
            self.scaler.fit(numerical_data)
            self.numerical_cols = numerical_data.columns.to_list()

        scaled_data = pd.DataFrame(
            self.scaler.transform(numerical_data),
            columns=self.numerical_cols,
        )

        return scaled_data

    def _inverse_ohe(self, data: pd.DataFrame) -> pd.DataFrame:
        """Inverse one-hot encode the data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to inverse one-hot encode.

        Returns
        -------
        undummified_df: pd.DataFrame
            The inverse one-hot encoded data.
        """
        df = data.select_dtypes(include="uint8")

        if df.empty or self.categorical_cols is None:
            return pd.DataFrame(index=data.index)

        cols2collapse = {
            item.split(self.prefix_sep)[0]: (self.prefix_sep in item)
            for item in df.columns
        }

        series_list = []
        for col, needs_to_collapse in cols2collapse.items():
            if needs_to_collapse & (col in self.categorical_cols):
                undummified = (
                    df.filter(like=col)
                    .idxmax(axis=1)
                    .apply(lambda x: x.split(self.prefix_sep, maxsplit=1)[1])
                    .rename(col)
                )
                series_list.append(undummified)
            else:
                series_list.append(df[col])
        undummified_df = pd.concat(series_list, axis=1)

        return undummified_df

    def _inverse_scale_numeric(self, data: pd.DataFrame) -> pd.DataFrame:
        """Inverse scale the numeric data.

        Parameters
        ----------
        data: pd.DataFrame
            The data to inverse scale.

        Returns
        -------
        pd.DataFrame
            The inverse scaled data.
        """
        df = data.select_dtypes(include="float64")

        if df.empty:
            return pd.DataFrame(index=data.index)

        return pd.DataFrame(
            self.scaler.inverse_transform(df), columns=df.columns
        )


class DataSet(Dataset):
    """The dataset class for the PyTorch dataloader."""

    def __init__(self, data: pd.DataFrame, labels: pd.DataFrame):
        """Initialize the dataset.
        Parameters
        ----------
        data: pd.DataFrame
            The data.
        labels: pd.DataFrame
            The labels.
        """
        self.data = data
        self.labels = labels

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, idx):
        x = torch.Tensor(self.data.iloc[idx])
        y = torch.Tensor(self.labels.iloc[idx])

        return x, y



"""LUCID."""

import warnings
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
import torch
import torch.nn as nn
import torch.optim as optim
from numpy.random import default_rng
from sklearn.base import TransformerMixin
from tqdm import tqdm

from canonical_sets.utils import safe_isinstance


class LUCID:
    """Gradient-based inverse design to generate canonical sets.

    This class generates a canonical set via inverse design and
    attributes the ``pd.DataFrame``  to ``results``.

    Attributes
    ----------
    results : pd.DataFrame
        A dataframe with the canonical inputs.
    results_processed: pd.DataFrame
        A dataframe with the processed canonical inputs.

    Examples
    --------
    >>> model = tf.keras.Model()
    >>> outputs = pd.DataFrame([[0, 1]], columns=["No", "Yes"])
    >>> example_data = train_data
    >>> lucid = LUCID(model, outputs, example_data)
    >>> lucid.results
    """

    results: pd.DataFrame
    results_processed: pd.DataFrame

    def __init__(
        self,
        model: Union[torch.nn.Module, tf.keras.Model],
        outputs: pd.DataFrame,
        example_data: pd.DataFrame,
        numb_of_samples: int = 100,
        numb_of_epochs: int = 200,
        lr: float = 0.1,
        low: float = -1,
        high: float = 1,
        seed: int = 1234,
        index: bool = True,
        extra_epoch: bool = True,
        one_hot_pre: bool = False,
        one_hot_post: bool = True,
        log_every_n: int = 0,
        prefix_sep: str = "+",
    ):
        """Initialize the inverse design.

        Parameters
        ----------
        model : torch.nn.Module or tf.keras.Model
            The trained model to use for inverse design.
        outputs : pd.DataFrame
            The outputs to use for inverse design. These are the
            targets/labels that have been used during training.
            For example, ``pd.DataFrame([[0, 1]], columns=["<=50K", ">50K"])``
            in the ``Adult`` data set.
        example_data : pd.DataFrame
            The example data to infer columns, dtypes, ...
            This is often (a part of) the training data itself,
            but can also be an artificial example.
        numb_of_samples : int
            The number of samples to generate. The default is 100.
        numb_of_epochs : int
            The number of epochs to train the model. The default is 200.
        lr : float
            The learning rate for the optimizer. The default is 0.1.
        low : float
            The lower bound for the random uniform distribution.
            The default is -1.
        high : float
            The upper bound for the random uniform distribution.
            The default is 1.
        seed : int
            The seed for the random number generator. The default is 1234.
        index : bool
            If True the sample and epoch numbers are used as indices in
            the results ``pd.DataFrame``. Otherwise they are just columns.
            The default is True.
        extra_epoch : bool
            If True an additional forward pass is run after the categorical
            features have been one-hot encoded (post-processed). The results
            are saved for the last sample as the ``numb_of_epochs`` + 1 epoch.
            If there are no categorical features the argument is ignored.
            The default is True.
        one_hot_pre : bool
            If True, the initial values for the categorical features are
            pre-processed to be one-hot. If there are no categorical
            features the argument is ignored. Note that the inverse
            design will start from this one-hot sample, hence the pre-
            process. If False, the inverse design will start from the
            randomly drawn initial vectors. The default is False.
        one_hot_post : bool
            If True, the values for the categorical features are
            post-processed to be one-hot. Note that the predictions during
            the inverse design are made with the original values of the
            categorical features and not with the post-processed values.
            To run an additional forward pass with the post-processed
            values check the ``extra_epoch`` argument. If there are no
            categorical features the argument is ignored. The default is True.
        log_every_n : int
            The number of epochs to log results. If 0, this argument is set
            equal to the ``numb_of_epochs`` argument which makes it a static
            analysis with only the start and end samples. The default is 0.
        prefix_sep : str
            The separator for the prefix of the column names. The one-hot
            encoded features are grouped via the prefix. To be safe, make
            sure that the prefix only appears as a prefix in the column
            names (i.e., avoid Categorical-category-name, and opt for
            Categorical+category-name instead). The default is "+".

        Raises
        ------
        ValueError
            If any columns are neither integer (one-hot encoded)
            or float (numerical).
        ValueError
            If the model is neither a torch.nn.Module or (tf.)keras.Model.
        """
        self.model = model
        self.outputs = outputs
        self.example_data = example_data
        self.numb_of_samples = numb_of_samples
        self.numb_of_epochs = numb_of_epochs
        self.lr = lr
        self.low = low
        self.high = high
        self.seed = seed
        self.index = index
        self.extra_epoch = extra_epoch
        self.one_hot_pre = one_hot_pre
        self.one_hot_post = one_hot_post
        self.log_every_n = log_every_n
        self.prefix_sep = prefix_sep

        if self.log_every_n == 0:
            self.log_every_n = self.numb_of_epochs

        self._processed = False

        # check whether all columns are either float or integer
        if len(self.example_data.columns) != len(
            self.example_data.select_dtypes(
                include=["float", "integer"]
            ).columns
        ):
            raise ValueError(
                (
                    "Some columns are neither integer (one-hot encoded) "
                    "or float (numerical)"
                )
            )

        # store column information
        self._input_columns = self.example_data.columns.to_list()
        self._numb_of_inputs = len(self._input_columns)

        self._output_columns = self.outputs.columns.to_list()

        self.numerical_cols = self.example_data.select_dtypes(
            include="float"
        ).columns.to_list()

        self.categories_cols = self.example_data.select_dtypes(
            include="integer"
        ).columns.to_list()

        self.categories = list(
            dict.fromkeys(
                [
                    item.split(self.prefix_sep)[0]
                    for item in self.categories_cols
                ]
            )
        )

        # generate random inputs
        self._rng = default_rng(seed=self.seed)
        self._inputs = self._rng.uniform(
            low=self.low,
            high=self.high,
            size=(self.numb_of_samples, self._numb_of_inputs),
        )

        # one-hot-encode categorical features (pre-processing, optional)
        if self.one_hot_pre and self.categories_cols:
            self._inputs = self._one_hot_encode(self._inputs)

        # initialize results
        self._columns = (
            ["sample", "epoch"] + self._output_columns + self._input_columns
        )
        self.results = pd.DataFrame(columns=self._columns)

        # run inverse design loop
        if safe_isinstance(self.model, "torch.nn.Module"):
            self._inverse_loop_pt()

        elif safe_isinstance(self.model, "keras.Model"):
            self._inverse_loop_tf()

        else:
            raise ValueError(
                "model must be a torch.nn.Module or (tf.)keras.Model."
            )

        # set types in results
        self.results["sample"] = self.results["sample"].astype(int)
        self.results["epoch"] = self.results["epoch"].astype(int)

        # save additional info
        self._min_epoch = 1
        self._max_epoch = self.numb_of_epochs - (
            self.numb_of_epochs % self.log_every_n
        )

        # set index
        if self.index:
            self.results.set_index(["sample", "epoch"], inplace=True)

        # one-hot-encode categorical features (post-processing, optional)
        if self.one_hot_post and self.categories_cols:
            self.results = self._one_hot_encode(self.results)

    def process_results(self, scaler: TransformerMixin = None) -> None:
        """Process the results by applying inverse scaler and one-hot
        encoding to categories.

        Parameters
        ----------
        scaler: sklearn.base.TransformerMixin, optional
            Any of the ``sklearn`` preprocessing modules.
            The default is None which means there is no transformation on
            numerical features.
        """
        if self._processed:
            warnings.warn("Results have already been processed before.")

        numbers = pd.DataFrame(index=self.results.index)
        dummies = pd.DataFrame(index=self.results.index)

        if self.numerical_cols and scaler:
            numbers = pd.DataFrame(
                scaler.inverse_transform(self.results[self.numerical_cols]),
                index=self.results.index,
                columns=self.numerical_cols,
            )

        elif self.numerical_cols:
            numbers = self.results[self.numerical_cols]

        if self.categories_cols:
            cols2collapse = {
                item.split(self.prefix_sep)[0]: (self.prefix_sep in item)
                for item in self.categories_cols
            }

            series_list = []
            for col, needs_to_collapse in cols2collapse.items():
                if needs_to_collapse:
                    undummified = (
                        self.results[self.categories_cols]
                        .filter(like=col)
                        .astype(float)
                        .idxmax(axis=1)
                        .apply(
                            lambda x: x.split(self.prefix_sep, maxsplit=1)[1]
                        )
                        .rename(col)
                    )
                    series_list.append(undummified)

                else:
                    series_list.append(self.results[col])
            dummies = pd.concat(series_list, axis=1)

        results_processed = numbers.join(dummies)

        self.results_processed = self.results[self._output_columns].join(
            results_processed
        )

        if self.index is False:
            self.results_processed = self.results[["sample", "epoch"]].join(
                self.results_processed
            )

        self._processed = True

    def plot(self, output: str) -> None:
        """Plot the outputs.

        Parameters
        ----------
        output: str
            The name of the output to plot.
        """
        if self.index:
            results = self.results

        else:
            results = self.results.set_index(["sample", "epoch"])

        plt.scatter(
            results.query(f"epoch == {self._min_epoch}")[
                output
            ].index.get_level_values("sample"),
            results.query(f"epoch == {self._min_epoch}")[output].values,
            color="blue",
        )

        plt.scatter(
            results.query(f"epoch == {self._max_epoch}")[
                output
            ].index.get_level_values("sample"),
            results.query(f"epoch == {self._max_epoch}")[output].values,
            color="red",
        )

        plt.scatter(
            results.query(f"epoch == {self._max_epoch + 1}")[
                output
            ].index.get_level_values("sample"),
            results.query(f"epoch == {self._max_epoch + 1}")[output].values,
            color="gray",
        )

        plt.title(f"{output}")

        plt.legend(
            [
                f"epoch {self._min_epoch}",
                f"epoch {self._max_epoch}",
                f"epoch {self._max_epoch + 1}",
            ]
        )

        plt.show()

    def hist(self, features: Union[str, List[str]]) -> None:
        """Plot the results for a given feature.

        Parameters
        ----------
        features: str or list of str
            The feature(s) to plot (either 1, 2, 3, 4, 6 or 8).

        Raises
        ------
        ValueError
            If the ``features`` are neither a string or a list of
            strings of size 2, 3, 4, 6 or 8.

        Note
        ----
        If the ``results`` are not yet processed, they will be
        with ``process_results``.
        """
        if not self._processed:
            self.process_results()

        if self.index:
            results = self.results_processed

        else:
            results = self.results_processed.set_index(["sample", "epoch"])

        if isinstance(features, str):
            plt.hist(
                results.query(f"epoch == {self._min_epoch}")[[features]],
                alpha=0.45,
                color="blue",
            )
            plt.hist(
                results.query(f"epoch == {self._max_epoch}")[[features]],
                alpha=0.45,
                color="red",
            )

            plt.title(f"{features}")

            plt.legend(
                [f"epoch {self._min_epoch}", f"epoch {self._max_epoch}"]
            )
            plt.show()

            return None

        if len(features) == 2:
            ncol = 2
            nrow = 1

        elif len(features) == 3:
            ncol = 3
            nrow = 1

        elif len(features) == 4:
            ncol = 2
            nrow = 2

        elif len(features) == 6:
            ncol = 3
            nrow = 2

        elif len(features) == 8:
            ncol = 4
            nrow = 2

        else:
            raise ValueError(
                "features must be either a string or list of "
                "size 2, 3, 4, 6 or 8."
            )

        fig, axs = plt.subplots(nrows=nrow, ncols=ncol)
        for feature, ax in zip(features, axs.ravel()):
            ax.hist(
                results.query(f"epoch == {self._min_epoch}")[[feature]],
                alpha=0.45,
                color="blue",
            )

            ax.hist(
                results.query(f"epoch == {self._max_epoch}")[[feature]],
                alpha=0.45,
                color="red",
            )

            ax.set_title(f"{feature}")
            ax.legend([f"epoch {self._min_epoch}", f"epoch {self._max_epoch}"])

    def _one_hot_encode(
        self, data: Union[pd.DataFrame, np.ndarray]
    ) -> Union[pd.DataFrame, np.ndarray]:
        """One-hot encode the data.

        This method takes a dataframe or a numpy array and one-hot encodes
        the categorical features. The return type is the same as the input
        type.

        Parameters
        ----------
        data: pd.DataFrame or np.ndarray
            The data where the categorical features should be one-hot encoded.

        Returns
        -------
        results: pd.DataFrame or np.ndarray
            The data where the categorical features are one-hot encoded.
        """
        if isinstance(data, np.ndarray):
            is_numpy_array = True
            data = pd.DataFrame(data, columns=self._input_columns)

        else:
            is_numpy_array = False

        results = data.drop(self.numerical_cols + self.categories_cols, axis=1)

        _results = pd.DataFrame(index=results.index)
        categories = self.categories.copy()

        for col in self._input_columns:

            if col in self.numerical_cols:
                _results[col] = data[col]

            elif col.split(self.prefix_sep)[0] in categories:
                categories.remove(col.split(self.prefix_sep)[0])

                filtered = (
                    data[self.categories_cols]
                    .filter(like=col.split(self.prefix_sep)[0])
                    .astype(float)
                )
                df = pd.DataFrame(
                    np.transpose(
                        np.where(filtered.T == filtered.T.max(), 1, 0)
                    ),
                    index=data.index,
                    columns=filtered.columns,
                )

                _results = pd.concat([_results, df], axis=1)

        _results = _results[self._input_columns]

        results = results.join(_results)

        if is_numpy_array:
            return results.to_numpy()

        return results

    def _inverse_loop_pt(self) -> None:
        """Inverse design loop for PyTorch."""

        # transform output
        y = torch.tensor(self.outputs.to_numpy(), dtype=torch.float32)

        # inverse design loop
        for i, x in enumerate(tqdm(self._inputs)):
            x = torch.tensor(x, dtype=torch.float32).unsqueeze(0)
            x.requires_grad_()

            criterion = nn.CrossEntropyLoss()
            optimizer = optim.SGD([x], lr=self.lr)

            for epoch in range(self.numb_of_epochs):
                sample = x.detach().numpy().flatten()

                optimizer.zero_grad()

                outputs = self.model(x)
                loss = criterion(outputs, y)

                loss.backward()
                optimizer.step()

                if (epoch + 1) % self.log_every_n == 0 or epoch == 0:
                    pred = outputs.detach().numpy().flatten()

                    row = np.concatenate(
                        [
                            np.array([(i + 1), (epoch + 1)]),
                            pred,
                            sample,
                        ]
                    )

                    self.results = pd.concat(
                        [
                            self.results,
                            pd.DataFrame([row], columns=self._columns),
                        ],
                        ignore_index=True,
                    )

            if self.categories_cols and self.extra_epoch:
                sample = self._one_hot_encode(np.expand_dims(sample, 0))

                output = self.model(torch.Tensor(sample))
                pred = output.detach().numpy().flatten()

                row = np.concatenate(
                    [np.array([(i + 1), (epoch + 2)]), pred, sample.squeeze()]
                )

                self.results = pd.concat(
                    [
                        self.results,
                        pd.DataFrame([row], columns=self._columns),
                    ],
                    ignore_index=True,
                )

    def _inverse_loop_tf(self) -> None:
        """Inverse design loop for Tensorflow."""
        # transform output
        y = self.outputs.to_numpy()

        # inverse design loop
        for i, x in enumerate(tqdm(self._inputs)):
            x = tf.Variable(np.expand_dims(x, axis=0))

            criterion = tf.keras.losses.BinaryCrossentropy()
            optimizer = tf.keras.optimizers.SGD(learning_rate=self.lr)

            for epoch in range(self.numb_of_epochs):
                sample = x.numpy()

                with tf.GradientTape() as tape:
                    tape.watch(x)
                    outputs = self.model(x, training=False)
                    loss = criterion(y, outputs)

                grads = tape.gradient(loss, x)
                optimizer.apply_gradients(zip([grads], [x]))

                if (epoch + 1) % self.log_every_n == 0 or epoch == 0:
                    pred = outputs.numpy()

                    row = np.concatenate(
                        [
                            np.array([[(i + 1), (epoch + 1)]]),
                            pred,
                            sample,
                        ],
                        axis=1,
                    )

                    self.results = pd.concat(
                        [
                            self.results,
                            pd.DataFrame(
                                [row.squeeze()], columns=self._columns
                            ),
                        ],
                        ignore_index=True,
                    )

            if self.categories_cols and self.extra_epoch:
                sample = self._one_hot_encode(sample)

                output = self.model(sample, training=False)
                pred = output.numpy()

                self.sample = sample
                self.output = output
                self.pred = pred

                row = np.concatenate(
                    [
                        np.array([(i + 1), (epoch + 2)]),
                        pred.squeeze(),
                        sample.squeeze(),
                    ]
                )

                self.results = pd.concat(
                    [
                        self.results,
                        pd.DataFrame([row], columns=self._columns),
                    ],
                    ignore_index=True,
                )



"""Sampler."""
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from ctgan.data_sampler import DataSampler
from ctgan.data_transformer import SpanInfo


class _Sampler(DataSampler):
    """Sampler wrapping ``DataSampler``.

    This class is based on the `DataSampler` class from the `ctgan`
    package. It has been modified to fix several bugs (see PRs on
    the `ctgan` GitHub page) and to allow for the extension of the
    conditional vector. Note that a part of the code and
    comments is identical to the original `DataSampler` class.
    """

    def __init__(
        self,
        data: np.ndarray,
        output_info: List[List[SpanInfo]],
        log_frequency: bool,
        conditions: Optional[np.ndarray] = None,
    ):
        """Initialize the sampler.

        Parameters
        ----------
        data : numpy.ndarray
            The data to be sampled.
        output_info : ctgan.data_transformer.SpanInfo
            The output information from the DataTransformer.
        log_frequency : bool
            Whether to use the log frequency.
        conditions : numpy.ndarray, optional
            The conditions to be sampled.
        """
        self._data = torch.from_numpy(data.astype("float32"))

        self._conditions = (
            torch.from_numpy(conditions.astype("float32"))
            if conditions is not None
            else None
        )

        def is_discrete_column(column_info):
            return (
                len(column_info) == 1
                and column_info[0].activation_fn == "softmax"
            )

        n_discrete_columns = sum(
            [
                1
                for column_info in output_info
                if is_discrete_column(column_info)
            ]
        )

        self._discrete_column_matrix_st = np.zeros(
            n_discrete_columns, dtype="int32"
        )

        # Store the row id for each category in each discrete column.
        # For example _rid_by_cat_cols[a][b] is a list of all rows with the
        # a-th discrete column equal value b.
        self._rid_by_cat_cols = []

        # Compute _rid_by_cat_cols
        st = 0
        for column_info in output_info:
            if is_discrete_column(column_info):
                span_info = column_info[0]
                ed = st + span_info.dim

                rid_by_cat = []
                for j in range(span_info.dim):
                    rid_by_cat.append(np.nonzero(data[:, st + j])[0])
                self._rid_by_cat_cols.append(rid_by_cat)
                st = ed
            else:
                st += sum([span_info.dim for span_info in column_info])
        assert st == data.shape[1]

        # Prepare an interval matrix for efficiently sample conditional vector
        max_category = max(
            [
                column_info[0].dim
                for column_info in output_info
                if is_discrete_column(column_info)
            ],
            default=0,
        )

        self._discrete_column_cond_st = np.zeros(
            n_discrete_columns, dtype="int32"
        )
        self._discrete_column_n_category = np.zeros(
            n_discrete_columns, dtype="int32"
        )
        self._discrete_column_category_prob = np.zeros(
            (n_discrete_columns, max_category)
        )
        self._n_discrete_columns = n_discrete_columns
        self._n_categories = sum(
            [
                column_info[0].dim
                for column_info in output_info
                if is_discrete_column(column_info)
            ]
        )

        st = 0
        current_id = 0
        current_cond_st = 0
        for column_info in output_info:
            if is_discrete_column(column_info):
                span_info = column_info[0]
                ed = st + span_info.dim
                category_freq = np.sum(data[:, st:ed], axis=0)
                if log_frequency:
                    category_freq = np.log(category_freq + 1)
                category_prob = category_freq / np.sum(category_freq)
                self._discrete_column_category_prob[
                    current_id, : span_info.dim
                ] = category_prob
                self._discrete_column_cond_st[current_id] = current_cond_st
                self._discrete_column_n_category[current_id] = span_info.dim
                self._discrete_column_matrix_st[current_id] = st

                current_cond_st += span_info.dim
                current_id += 1
                st = ed
            else:
                st += sum([span_info.dim for span_info in column_info])

    def sample_original_condvec(
        self, batch: int
    ) -> Union[np.ndarray, Tuple[np.ndarray, torch.Tensor]]:
        """Generate the conditional vector using the original frequency.

        Parameters
        ----------
        batch : int
            The batch size.

        Returns
        -------
        np.ndarray, torch.Tensor, optional
            The conditional vector and (optionally) extra conditions.
        """
        cond = np.zeros((batch, self._n_categories), dtype="float32")

        if self._conditions is not None:
            ids = []

        for i in range(batch):
            row_idx = np.random.randint(0, len(self._data))
            col_idx = np.random.randint(0, self._n_discrete_columns)
            matrix_st = self._discrete_column_matrix_st[col_idx]
            matrix_ed = matrix_st + self._discrete_column_n_category[col_idx]
            pick = np.argmax(self._data[row_idx, matrix_st:matrix_ed])
            cond[i, pick + self._discrete_column_cond_st[col_idx]] = 1

            if self._conditions is not None:
                ids.append(row_idx)

        if self._conditions is not None:
            return cond, self._conditions[ids]
        else:
            return cond

    def sample_data(
        self, n: int, col: Optional[int] = None, opt: Optional[int] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Sample data from original training data for the conditional vector.

        Parameters
        ----------
        n : int
            The number of samples.
        col : int, optional
            The column to be sampled.
        opt : int, optional
            The category to be sampled.

        Returns
        -------
        torch.Tensor
            The conditional vector and (optionally) extra conditions.
        """
        if col is None and opt is None:
            idx = np.random.randint(len(self._data), size=n)

            if self._conditions is None:
                return self._data[idx, :]

            else:
                return self._data[idx, :], self._conditions[idx, :]

        else:
            idx_list = []
            for c, o in zip(col, opt):  # type: ignore
                idx_list.append(np.random.choice(self._rid_by_cat_cols[c][o]))

            if self._conditions is None:
                return self._data[idx_list, :]

            else:
                return self._data[idx_list, :], self._conditions[idx_list, :]

    def generate_cond_from_condition_column_info(
        self, condition_infos: List[Dict[str, int]], batch: int
    ) -> np.ndarray:
        """Generate the condition vector.

        Parameters
        ----------
        condition_infos : List[Dict[str, int]]
            The condition column information.
        batch : int
            The batch size.

        Returns
        -------
        numpy.ndarray
            The condition vector.
        """
        vec = np.zeros((batch, self._n_categories), dtype="float32")

        for condition_info in condition_infos:
            id_ = self._discrete_column_cond_st[
                condition_info["discrete_column_id"]
            ]
            id_ += condition_info["value_id"]
            vec[:, id_] = 1

        return vec



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['canonical_sets',
 'canonical_sets.data',
 'canonical_sets.gan',
 'canonical_sets.models']

package_data = \
{'': ['*']}

install_requires = \
['ctgan>=0.6.0',
 'matplotlib>=3.5.2',
 'numpy>=1.22.3',
 'pandas>=1.4.2',
 'scikit-learn>=1.1.1',
 'seaborn>=0.12.0',
 'tensorflow>=2.9.1,<2.11.0',
 'torch>=1.11.0,<1.13.0',
 'tqdm>=4.64.0']

setup_kwargs = {
    'name': 'canonical-sets',
    'version': '0.0.2',
    'description': 'Exposing Algorithmic Bias with Canonical Sets.',
    'long_description': '.. |nbsp| unicode:: U+00A0 .. NO-BREAK SPACE\n\n.. |pic1| image:: https://img.shields.io/badge/python-3.8%20%7C%203.9-blue\n.. |pic2| image:: https://img.shields.io/github/license/mashape/apistatus.svg\n.. |pic3| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n.. |pic4| image:: https://img.shields.io/badge/%20type_checker-mypy-%231674b1?style=flat\n.. |pic5| image:: https://img.shields.io/badge/platform-windows%20%7C%20linux%20%7C%20macos-lightgrey\n.. |pic6| image:: https://github.com/Integrated-Intelligence-Lab/canonical_sets/actions/workflows/testing.yml/badge.svg\n.. |pic7| image:: https://img.shields.io/readthedocs/canonical_sets\n.. |pic8| image:: https://img.shields.io/pypi/v/canonical_sets\n\n.. _canonical_sets: https://github.com/Integrated-Intelligence-Lab/canonical_sets/tree/main/canonical_sets\n.. _examples: https://github.com/Integrated-Intelligence-Lab/canonical_sets/tree/main/examples\n.. _contribute: https://github.com/Integrated-Intelligence-Lab/canonical_sets/blob/main/CONTRIBUTING.rst\n.. _documentation: https://canonical-sets.readthedocs.io/en/latest/\n.. _LUCID: https://responsibledecisionmaking.github.io/assets/pdf/papers/21.pdf\n.. _LUCIDGAN: https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4289597\n\n.. _Twitter: https://twitter.com/DataLabBE\n.. _website: https://data.research.vub.be/\n.. _papers: https://researchportal.vub.be/en/organisations/data-analytics-laboratory/publications/\n\n.. _ctgan: https://github.com/sdv-dev/CTGAN\n.. _ctganbugs: https://github.com/sdv-dev/CTGAN/pulls/AndresAlgaba\n\n\nCanonical sets \n==============\n\n|pic2| |nbsp| |pic5| |nbsp| |pic1| |nbsp| |pic8|\n\n|pic6| |nbsp| |pic7| |nbsp| |pic3| |nbsp| |pic4|\n\nAI systems can create, propagate, support, and automate bias in decision-making processes. To mitigate biased decisions,\nwe both need to understand the origin of the bias and define what it means for an algorithm to make fair decisions.\nBy Locating Unfairness through Canonical Inverse Design (LUCID), we generate a canonical set that shows the desired inputs\nfor a model given a preferred output. The canonical set reveals the model\'s internal logic and exposes potential unethical\nbiases by repeatedly interrogating the decision-making process.\n\nLUCID-GAN extends on LUCID by generating canonical inputs via a conditional generative model instead of\ngradient-based inverse design. LUCID-GAN generates canonical inputs conditional on the predictions of the model under\nfairness evaluation. LUCID-GAN has several benefits, including that it applies to non-differentiable models, ensures\nthat a canonical set consists of realistic inputs, and allows us to assess indirect discrimination and explicitly\ncheck for intersectional unfairness.\n\nRead our paper on `LUCID`_ and `LUCID-GAN`_ for more details, or check out the `documentation`_.\n\nWe encourage everyone to `contribute`_ to this project by submitting an issue or a pull request!\n\n\nInstallation\n------------\n\nInstall ``canonical_sets`` from PyPi.\n\n.. code-block:: bash\n\n    pip install canonical_sets\n\nFor development install, see `contribute`_. You can also check the `documentation`_.\n\n\nUsage\n-----\n\nLUCID\n~~~~~\n\n``LUCID`` can be used for the gradient-based inverse design to generate canonical sets, and is available for both\n``PyTorch`` and ``Tensorflow`` models. It only requires a model, a preferred output, and an example input\n(which is often a part of the training data). The results are stored in a ``pd.DataFrame``, and can be accessed by\ncalling ``results``. It\'s fully customizable, but can also be used out-of-the-box for a wide range of\napplications by using its default settings:\n\n.. code-block:: python\n\n    from canonical_sets import LUCID\n\n    lucid = LUCID(model, outputs, example_data)\n    lucid.results.head()\n\nLUCID-GAN\n~~~~~~~~~\n\n``LUCIDGAN`` generates canonical sets by using conditional generative models (GANs). This approach has several benefits,\nincluding that it applies to non--differentiable models, ensures that a canonical set consists of realistic inputs,\nand allows us to assess indirect discrimination and explicitly check for intersectional unfairness. LUCID-GAN only\nrequires the input and predictions of a black-box model. It\'s fully customizable, but can also be used out-of-the-box\nfor a wide range of applications by using its default settings:\n\n.. code-block:: python\n\n    from canonical_sets import LUCIDGAN\n\n    lucidgan = LUCIDGAN()\n    lucidgan.fit(data, predictions)\n    samples = lucidgan.sample(100)\n    samples.head()\n\nFor detailed examples see `examples`_ and for the source code see `canonical_sets`_. For ``LUCID``, we advice to start with either the\n``tensorflow`` or ``pytorch`` example, and then the advanced example. For ``LUCIDGAN``, you can replicate the experiments from the paper\nwith the ``GAN_adult`` and ``GAN_compas`` examples. Note that the results might slightly differ due to the randomness in generating the\nsamples. You can also check the `documentation`_ for more details. If you have any remaining questions, feel free to submit an issue or PR!\n\n\nOutput-based group metrics\n~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nMost group fairness notions focus on the equality of outcome by computing statistical parity metrics on a model\'s output.\nThe two most prominent examples of these statistical output-based metrics are Demographic Parity (DP) and Equality Of Opportunity (EOP).\nIn DP, we compare the Positivity Rate (PR) of the subpopulations under fairness evaluation, and in EOP, we compare the True Positive Rate (TPR).\nThe choice between DP and EOP depends on the underlying assumptions and worldview of the evaluator.\nThe ``Metrics`` class allows you to compute these metrics for binary classification tasks given the predictions and ground truth:\n\n.. code-block:: python\n\n    from canonical_sets.group import Metrics\n\n    metrics = Metrics(preds, targets)\n    metrics.metrics\n\n\nData\n----\n\n``canonical_sets`` contains some functionality to easily access commonly used data sets in the fairness literature:\n\n.. code-block:: python\n\n    from canonical_sets import Adult, Compas\n\n    adult = Adult()\n    adult.train_data.head()\n\n    compas = Compas()\n    compas.train_data.head()\n\nThe default settings can be customized to change the pre-processing, splitting, etc. See `examples`_  for details.\nYou can also check the `documentation`_.\n\n\nCommunity\n---------\n\nIf you are interested in cross-disciplinary research related to machine learning, feel free to:\n\n* Follow DataLab on `Twitter`_.\n* Check the `website`_.\n* Read our `papers`_.\n\n\nDisclaimer\n----------\n\nThe package and the code is provided "as-is" and there is NO WARRANTY of any kind. \nUse it only if the content and output files make sense to you.\n\nCurrently some dependencies of the package do not support the Apple M1 and M2 chips.\nWe will offer support asap.\n\n\nAcknowledgements\n----------------\n\nThis project benefited from financial support from Innoviris.\n\n``LUCIDGAN`` is based on the ``CTGAN`` class from the `ctgan`_ package. It has been extended to fix\nseveral bugs (see my PRs on the `ctganbugs`_ GitHub page) and to allow for the extension of the conditional\nvector. Note that a part of the code and comments is identical to the original ``CTGAN`` class.\n\n\nCitation\n--------\n\n.. code-block:: none\n\n    @inproceedings{mazijn_lucid_2023,\n      title={{LUCID: Exposing Algorithmic Bias through Inverse Design}},\n      author={Mazijn, Carmen and Prunkl, Carina and Algaba, Andres and Danckaert, Jan and Ginis, Vincent},\n      booktitle={Thirty-Seventh AAAI Conference on Artificial Intelligence (accepted)},\n      year={2023},\n    }\n\n    @article{algaba_lucidgan_2022,\n      title={{LUCID-GAN: Conditional Generative Models to Locate Unfairness}},\n      author={Algaba, Andres and Mazijn, Carmen and Prunkl, Carina and Danckaert, Jan and Ginis, Vincent},\n      year={2022},\n      journal={Working paper}\n    }\n\n',
    'author': 'Integrated Intelligence Lab',
    'author_email': None,
    'maintainer': 'Andres Algaba',
    'maintainer_email': 'andres.algaba@vub.be',
    'url': 'https://data.research.vub.be/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
